#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libspe2.h>
#include <pthread.h>
#include "nbody.h"
#include "spe_poor.h"

// 
// ̃vÓATCYŜSPEɏ؂邱Ƃ肵Ă
// 5 * nplots < 256K
#define SPE_ALIGN	16
#define NSPE	6

typedef struct spe_thread_args {
	spe_context_ptr_t spe;
	spe_stop_info_t* spe_si;
	unsigned int entry;
	spe_args_t* speargs;
} spe_thread_args_t;

typedef struct step_args {
	spe_thread_args_t* targs;
} step_args_t;

static void do_step(step_args_t* args);
static void *run_spe(void* args);

int main(int argc, char** argv) {
	point_t *plots;
	state_t st;
	char *pgmname, *inputname, *outputname;
	float *mass, *pos_x, *pos_y, *vel_x, *vel_y;
	void * tmp_mem;
	int aligned_buffer_size, aligned_nplots;
	int i, j, data_size, tmp;
	int nstep, offset = 0;

	spe_context_ptr_t spe[NSPE];
	spe_stop_info_t spe_si[NSPE];
	spe_program_handle_t *pgmhdl;
	spe_args_t speargs[NSPE] __attribute__((aligned(16)));
	spe_thread_args_t stargs[NSPE];
	int n_aligned_block, assigned_block;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	step_args_t stepargs;

	pgmname = argv[1];
	inputname = argv[2];
	outputname = argv[3];
	nstep = atoi(argv[4]);

	printf("PGM: %s\n", pgmname);

	// t@Cǂݍ
	init_plots_map(inputname, &st, &plots);

	// 16bytẽubNAʐMobt@TCYZo
	data_size = sizeof(float) * st.nplots;
	tmp = (int)(data_size / SPE_ALIGN);
	n_aligned_block = tmp + ((data_size - tmp * SPE_ALIGN > 0) & 0x1);
	aligned_buffer_size = n_aligned_block * SPE_ALIGN;
	aligned_nplots = aligned_buffer_size / sizeof(float);

	// ]œK̂߂̃f[^ۑꏊ 
	// malloc  8byte align Ȃ̂ 16byte align ɂ posix_memalign KvI
	if (posix_memalign(&tmp_mem, SPE_ALIGN, aligned_buffer_size) != 0) {
		perror("posix_memalign mass");
	}
	mass = (float *) tmp_mem;
	if (posix_memalign(&tmp_mem, SPE_ALIGN, aligned_buffer_size) != 0) {
		perror("posix_memalign pos_x");
	}
	pos_x = (float *) tmp_mem;
	if (posix_memalign(&tmp_mem, SPE_ALIGN, aligned_buffer_size) != 0) {
		perror("posix_memalign pos_y");
	}
	pos_y = (float *) tmp_mem;
	if (posix_memalign(&tmp_mem, SPE_ALIGN, aligned_buffer_size) != 0) {
		perror("posix_memalign vel_x");
	}
	vel_x = (float *) tmp_mem;
	if (posix_memalign(&tmp_mem, SPE_ALIGN, aligned_buffer_size) != 0) {
		perror("posix_memalign vel_y");
	}
	vel_y = (float *) tmp_mem;

	for (i = 0; i < st.nplots; i++) {
		mass[i] = plots[i].m;
		pos_x[i] = plots[i].x;
		pos_y[i] = plots[i].y;
		vel_x[i] = plots[i].vx;
		vel_y[i] = plots[i].vy;
	}
	for (i = st.nplots; i < aligned_nplots; i++) {
		mass[i] = 0.;
		pos_x[i] = 0.;
		pos_y[i] = 0.;
		vel_x[i] = 0.;
		vel_y[i] = 0.;
	}

	// SPE PROGRAM nh쐬
	if ((pgmhdl = spe_image_open(pgmname)) == NULL) {
		perror("spe_image_open");
		exit(EXIT_FAILURE);
	}

	for (i = 0; i < NSPE; i++) {		
		// SPE CONTEXT ̍쐬ASPE PROGRAM ̃[h
		if (!(spe[i] = spe_context_create(0, NULL))) {
			perror("spe_context_create");
			exit(EXIT_FAILURE);
		}
		if ((spe_program_load(spe[i], pgmhdl))) {
			perror("spe_program_load");
			exit(EXIT_FAILURE);
		}

		// THREAD ARGS 쐬
		stargs[i].entry = entry;
		stargs[i].spe = spe[i];
		stargs[i].spe_si = &(spe_si[i]);
		stargs[i].speargs = &(speargs[i]);

		// SPE ARGS ̍쐬		
		speargs[i].mass = (unsigned long)mass;
		speargs[i].pos_x = (unsigned long)pos_x;
		speargs[i].pos_y = (unsigned long)pos_y;
		speargs[i].vel_x = (unsigned long)vel_x;
		speargs[i].vel_y = (unsigned long)vel_y;

		speargs[i].assigned_id_min = offset;
		assigned_block = (int)((n_aligned_block / NSPE ) + (((n_aligned_block
				% NSPE) > i) & 0x1));
		offset = offset + assigned_block * SPE_ALIGN / sizeof(float);
		speargs[i].assigned_id_max = offset;
		speargs[i].nplots = st.nplots;
		speargs[i].aligned_buffer_size = aligned_buffer_size;

		speargs[i].gc = st.gc;
		speargs[i].delta_t = st.delta_t;
	}

	// STEP ARGS 쐬
	stepargs.targs = stargs;
	for (j = 0; j < nstep; j++) {
		do_step(&stepargs);
	}

	// SPE CONTEXT j
		
	for (i = 0; i < NSPE; i++) {
		tmp = spe_context_destroy(spe[i]);
		if (tmp) {
			perror("spe_context_destroy");
			exit(EXIT_FAILURE);
		}
	}

	// f[^
	for (i = 0; i < st.nplots; i++) {
		plots[i].m = mass[i];
		plots[i].x = pos_x[i];
		plots[i].y = pos_y[i];
		plots[i].vx = vel_x[i];
		plots[i].vy = vel_y[i];
	}
	free(mass);
	free(pos_x);
	free(pos_y);
	free(vel_x);
	free(vel_y);

	write_plots_map(outputname, &st, plots);
	fini_plots_map(plots);

	return 1;
}

// ӁFXbhv[gׂʓ|Ȃ̂ŖĂ
// XbhƂ̒ʐMKvȂ̂ł߂ǂ
static void do_step(step_args_t* args) {
	static int s = 0;
	int i, tmp;
	pthread_t thread[NSPE];
	for (i = 0; i < NSPE; i++) {
		if (pthread_create(&thread[i], NULL, run_spe, &(args->targs[i])) != 0) {
			perror("pthread_create");
			exit(EXIT_FAILURE);
		}
	}

	// SPE 
	for (i = 0; i < NSPE; i++) {
		pthread_join(thread[i], NULL);
		// Ggʒu
		args->targs[i].entry =  SPE_DEFAULT_ENTRY;
		
/*		tmp = spe_context_destroy(args->targs[i].spe);
		if (tmp) {
			perror("spe_context_destroy");
			exit(EXIT_FAILURE);
		}
*/	}
	s++;
}

static void *run_spe(void* args) {
	int ret;
	spe_thread_args_t *stargs = (spe_thread_args_t*) args;

	ret = spe_context_run(stargs->spe, &(stargs->entry), 0, stargs->speargs,
			NULL, stargs->spe_si);
	if (ret < 0) {
		perror("spe_context_run");
		return NULL;
	}
	return NULL;
}
